"""
Tests for schema validation.

This module contains tests for the Pydantic schemas used for validation.
"""

import pytest
from pydantic import ValidationError

from umbrel_linter.schemas.umbrel_app import UmbrelAppManifest, UmbrelAppStoreManifest


class TestUmbrelAppManifest:
    """Test UmbrelAppManifest schema."""
    
    def test_valid_manifest(self):
        """Test valid manifest creation."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        manifest = UmbrelAppManifest(**manifest_data)
        assert manifest.id == "test-app"
        assert manifest.name == "Test App"
        assert manifest.port == 3000
    
    def test_invalid_manifest_version(self):
        """Test invalid manifest version."""
        manifest_data = {
            "manifest_version": 2.0,  # Invalid version
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "manifest_version" in str(exc_info.value)
    
    def test_invalid_app_id(self):
        """Test invalid app ID."""
        manifest_data = {
            "manifest_version": 1,
            "id": "umbrel-app-store-test",  # Invalid ID
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "umbrel-app-store" in str(exc_info.value)
    
    def test_invalid_tagline(self):
        """Test invalid tagline."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application.",  # Ends with period
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "period" in str(exc_info.value)
    
    def test_invalid_category(self):
        """Test invalid category."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "invalid_category",  # Invalid category
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "category" in str(exc_info.value)
    
    def test_invalid_port(self):
        """Test invalid port."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 70000,  # Invalid port
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "port" in str(exc_info.value)
    
    def test_self_dependency(self):
        """Test self-dependency validation."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app",
            "dependencies": ["test-app"]  # Self-dependency
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppManifest(**manifest_data)
        
        assert "Dependencies" in str(exc_info.value)
    
    def test_optional_fields(self):
        """Test optional fields."""
        manifest_data = {
            "manifest_version": 1,
            "id": "test-app",
            "name": "Test App",
            "tagline": "A test application",
            "category": "files",
            "version": "1.0.0",
            "port": 3000,
            "description": "This is a test application",
            "developer": "Test Developer",
            "submitter": "Test Submitter",
            "submission": "https://github.com/user/repo/pull/123",
            "support": "https://example.com/support",
            "website": "https://example.com",
            "path": "/app",
            "icon": "https://example.com/icon.png",
            "gallery": ["https://example.com/image1.png"],
            "release_notes": "Version 1.0.0",
            "dependencies": ["other-app"],
            "permissions": ["STORAGE_DOWNLOADS"],
            "default_username": "admin",
            "default_password": "password",
            "deterministic_password": True,
            "optimized_for_umbrel_home": True,
            "tor_only": False,
            "install_size": 1024,
            "default_shell": "/bin/bash",
            "backup_ignore": ["*.log", "*.tmp"],
            "repo": "https://github.com/user/repo"
        }
        
        manifest = UmbrelAppManifest(**manifest_data)
        assert manifest.icon == "https://example.com/icon.png"
        assert manifest.gallery == ["https://example.com/image1.png"]
        assert manifest.release_notes == "Version 1.0.0"
        assert manifest.dependencies == ["other-app"]
        assert manifest.permissions == ["STORAGE_DOWNLOADS"]
        assert manifest.default_username == "admin"
        assert manifest.default_password == "password"
        assert manifest.deterministic_password is True
        assert manifest.optimized_for_umbrel_home is True
        assert manifest.tor_only is False
        assert manifest.install_size == 1024
        assert manifest.default_shell == "/bin/bash"
        assert manifest.backup_ignore == ["*.log", "*.tmp"]
        assert manifest.repo == "https://github.com/user/repo"


class TestUmbrelAppStoreManifest:
    """Test UmbrelAppStoreManifest schema."""
    
    def test_valid_store_manifest(self):
        """Test valid store manifest creation."""
        store_data = {
            "id": "my-app-store",
            "name": "My App Store"
        }
        
        store = UmbrelAppStoreManifest(**store_data)
        assert store.id == "my-app-store"
        assert store.name == "My App Store"
    
    def test_invalid_store_id_reserved(self):
        """Test invalid store ID with reserved prefix."""
        store_data = {
            "id": "umbrel-app-store-test",  # Reserved prefix
            "name": "Test Store"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppStoreManifest(**store_data)
        
        assert "umbrel-app-store" in str(exc_info.value)
    
    def test_invalid_store_id_format(self):
        """Test invalid store ID format."""
        store_data = {
            "id": "My_App_Store",  # Invalid format (uppercase, underscores)
            "name": "Test Store"
        }
        
        with pytest.raises(ValidationError) as exc_info:
            UmbrelAppStoreManifest(**store_data)
        
        assert "alphabets" in str(exc_info.value)
    
    def test_valid_store_id_with_dashes(self):
        """Test valid store ID with dashes."""
        store_data = {
            "id": "my-awesome-app-store",
            "name": "My Awesome App Store"
        }
        
        store = UmbrelAppStoreManifest(**store_data)
        assert store.id == "my-awesome-app-store"
        assert store.name == "My Awesome App Store"



"""
Command-line interface for the Umbrel linter.

This module provides a user-friendly CLI for running the linter
with various options and configurations.
"""

from __future__ import annotations

from pathlib import Path

import typer
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from ..core.linter import UmbrelLinter
from ..core.models import AppStoreType, LinterConfig, LintingContext, Severity
from .fixer import apply_fixes

# Initialize CLI app
app = typer.Typer(
    name="ulint",
    help="A linter for Umbrel app stores and applications",
    add_completion=False,
)

# Initialize console for rich output
console = Console()


@app.command()
def lint(
    path: str = typer.Argument(..., help="Path to the directory to lint"),
    app_id: str | None = typer.Option(
        None, "--app", "-a", help="Specific app ID to lint"
    ),
    log_level: str = typer.Option(
        "warning", "--log-level", "-l", help="Log level (error, warning, info)"
    ),
    strict: bool = typer.Option(
        False, "--strict", "-s", help="Treat warnings as errors"
    ),
    skip_architectures: bool = typer.Option(
        False, "--skip-architectures", help="Skip Docker image architecture checks"
    ),
    new_submission: bool = typer.Option(
        False, "--new-submission", help="This is a new app submission"
    ),
    pr_url: str | None = typer.Option(
        None, "--pr-url", help="Pull request URL for new submissions"
    ),
    store_type: str = typer.Option(
        "community", "--store-type", help="App store type (official, community)"
    ),
    output_format: str = typer.Option(
        "rich", "--format", "-f", help="Output format (rich, json, plain)"
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output"),
    fix: bool = typer.Option(False, "--fix", help="Automatically fix safe issues"),
    fix_secure: bool = typer.Option(
        False,
        "--fix-secure",
        help="Apply additional secure fixes (e.g., set container user)",
    ),
) -> None:
    """
    Lint Umbrel applications and app stores.

    This command performs comprehensive linting of Umbrel applications,
    including YAML validation, Docker Compose validation, and security checks.
    """
    try:
        # Parse and validate arguments
        directory = Path(path).resolve()
        if not directory.exists():
            console.print(f"[red]Error:[/red] Directory '{path}' does not exist")
            raise typer.Exit(1)

        if not directory.is_dir():
            console.print(f"[red]Error:[/red] '{path}' is not a directory")
            raise typer.Exit(1)

        # Parse log level
        try:
            severity = Severity(log_level.lower())
        except ValueError:
            console.print(
                f"[red]Error:[/red] Invalid log level '{log_level}'. Must be one of: error, warning, info"
            )
            raise typer.Exit(1)

        # Parse store type
        try:
            app_store_type = AppStoreType(store_type.lower())
        except ValueError:
            console.print(
                f"[red]Error:[/red] Invalid store type '{store_type}'. Must be one of: official, community"
            )
            raise typer.Exit(1)

        # Create configuration
        config = LinterConfig(
            check_image_architectures=(not skip_architectures),
            log_level=severity,
            strict_mode=strict,
        )

        # Create context
        context = LintingContext(
            app_id=app_id,
            app_store_type=app_store_type,
            is_new_submission=new_submission,
            pull_request_url=pr_url,
        )

        # Initialize linter
        linter = UmbrelLinter(config)

        # Run linting
        import asyncio

        if app_id:
            console.print(f"[blue]Linting app:[/blue] {app_id}")
            result = asyncio.run(linter.lint_app(directory, app_id, context))
        else:
            console.print(f"[blue]Linting directory:[/blue] {directory}")
            result = asyncio.run(linter.lint_all_apps(directory, context))

        # Display results
        if output_format == "json":
            _display_json_output(result)
        elif output_format == "plain":
            _display_plain_output(result, severity)
        else:
            _display_rich_output(result, severity, verbose)

        # Offer auto-fix guidance when issues are present
        fixable_ids = {
            "empty_app_data_directory",
            "missing_file_or_directory",
            "invalid_yaml_boolean_value",
            "invalid_restart_policy",
            "invalid_tagline",
            "invalid_container_user",
        }
        if any(err.id in fixable_ids for err in result.errors):
            console.print(
                "\n[dim]Tip:[/dim] You can auto-fix many issues with:\n  [bold]umbrel-linter lint[/bold] [cyan]<path>[/cyan] --fix\n  [bold]umbrel-linter lint[/bold] [cyan]<path>[/cyan] --fix --fix-secure"
            )

        # Apply fixes if requested
        if fix or fix_secure:
            apply_fixes(directory, app_id, result, fix_secure, console)
            console.print("\n[green]Auto-fix completed. Re-run lint to verify.[/green]")

        # Exit with appropriate code
        # Only exit with error code if there are actual errors (not just info messages)
        if result.has_errors() or (strict and result.has_warnings()):
            raise typer.Exit(1)
        # Success case - no need to raise typer.Exit(0), just return

    except KeyboardInterrupt:
        console.print("\n[yellow]Linting interrupted by user[/yellow]")
        raise typer.Exit(1)
    except typer.Exit:
        # Re-raise typer.Exit to preserve exit codes
        raise
    except Exception as e:
        console.print(f"[red]Unexpected error:[/red] {e}")
        if verbose:
            console.print_exception()
        raise typer.Exit(1)


def _display_rich_output(result, severity: Severity, verbose: bool) -> None:
    """Display results using rich formatting."""

    # Summary
    error_count = result.total_errors
    warning_count = result.total_warnings
    info_count = result.total_info

    if result.success and warning_count == 0 and info_count == 0:
        console.print(
            Panel(
                "[green]✓ No linting errors found! 🎉[/green]",
                title="Linting Complete",
                border_style="green",
            )
        )
    elif result.success:
        # No errors, but has warnings or info messages
        summary_parts = []
        if warning_count > 0:
            summary_parts.append(f"[yellow]{warning_count} warnings[/yellow]")
        if info_count > 0:
            summary_parts.append(f"[blue]{info_count} info messages[/blue]")

        summary_text = (
            f"[green]✓ Linting passed with {', '.join(summary_parts)}[/green]"
        )
        console.print(
            Panel(
                summary_text,
                title="Linting Complete",
                border_style="green",
            )
        )
    else:
        summary_text = f"[red]✗ Linting failed with {error_count} errors[/red]"
        if warning_count > 0:
            summary_text += f", [yellow]{warning_count} warnings[/yellow]"
        if info_count > 0:
            summary_text += f", [blue]{info_count} info messages[/blue]"

        console.print(
            Panel(
                summary_text,
                title="Linting Complete",
                border_style="red",
            )
        )

    # Filter errors by log level
    filtered_errors = []
    for error in result.errors:
        if (
            (severity == Severity.ERROR and error.severity == Severity.ERROR)
            or (
                severity == Severity.WARNING
                and error.severity in [Severity.ERROR, Severity.WARNING]
            )
            or severity == Severity.INFO
        ):
            filtered_errors.append(error)

    if not filtered_errors:
        return

    # Group errors by file
    errors_by_file = {}
    for error in filtered_errors:
        if error.file not in errors_by_file:
            errors_by_file[error.file] = []
        errors_by_file[error.file].append(error)

    # Display errors
    for file_path, file_errors in errors_by_file.items():
        console.print(f"\n[bold]{file_path}[/bold]")

        for error in file_errors:
            # Color based on severity
            if error.severity == Severity.ERROR:
                severity_color = "red"
                severity_icon = "✗"
            elif error.severity == Severity.WARNING:
                severity_color = "yellow"
                severity_icon = "⚠"
            else:
                severity_color = "blue"
                severity_icon = "ℹ"

            # Create error text
            error_text = Text()
            error_text.append(f"{severity_icon} ", style=severity_color)
            error_text.append(f"{error.severity.upper()} ", style=severity_color)
            error_text.append(f"{error.title}", style="bold")

            if error.properties_path:
                error_text.append(f" ({error.properties_path})", style="dim")

            console.print(error_text)
            console.print(f"  {error.message}", style="dim")

            if verbose and error.line:
                console.print(
                    f"  Line {error.line.start}-{error.line.end}", style="dim"
                )


def _display_plain_output(result, severity: Severity) -> None:
    """Display results in plain text format."""

    if result.success:
        print("No linting errors found!")
        return

    # Filter errors by log level
    filtered_errors = []
    for error in result.errors:
        if (
            (severity == Severity.ERROR and error.severity == Severity.ERROR)
            or (
                severity == Severity.WARNING
                and error.severity in [Severity.ERROR, Severity.WARNING]
            )
            or severity == Severity.INFO
        ):
            filtered_errors.append(error)

    for error in filtered_errors:
        print(f"{error.severity.upper()}: {error.title}")
        print(f"  File: {error.file}")
        if error.properties_path:
            print(f"  Path: {error.properties_path}")
        print(f"  Message: {error.message}")
        if error.line:
            print(f"  Line: {error.line.start}-{error.line.end}")
        print()


def _display_json_output(result) -> None:
    """Display results in JSON format."""
    import json

    output = {
        "success": result.success,
        "total_errors": result.total_errors,
        "total_warnings": result.total_warnings,
        "total_info": result.total_info,
        "errors": [
            {
                "id": error.id,
                "severity": error.severity,
                "title": error.title,
                "message": error.message,
                "file": error.file,
                "properties_path": error.properties_path,
                "line": (
                    {
                        "start": error.line.start,
                        "end": error.line.end,
                    }
                    if error.line
                    else None
                ),
                "column": (
                    {
                        "start": error.column.start,
                        "end": error.column.end,
                    }
                    if error.column
                    else None
                ),
            }
            for error in result.errors
        ],
    }

    print(json.dumps(output, indent=2))


@app.command()
def version() -> None:
    """Show version information."""
    from .. import __version__

    console.print(f"umbrel-linter version {__version__}")


@app.command()
def config() -> None:
    """Show configuration information."""
    config = LinterConfig()

    table = Table(title="Linter Configuration")
    table.add_column("Setting", style="cyan")
    table.add_column("Value", style="green")

    table.add_row("Check Image Architectures", str(config.check_image_architectures))
    table.add_row("Log Level", config.log_level.value)
    table.add_row("Strict Mode", str(config.strict_mode))
    table.add_row("Ignore Patterns", ", ".join(config.ignore_patterns) or "None")

    console.print(table)


if __name__ == "__main__":
    app()



from __future__ import annotations

from pathlib import Path
from typing import Iterable

import yaml
from rich.console import Console

from ..core.models import LintingError
from ..utils.filesystem import write_file_safely


def apply_fixes(root: Path, app_id: str | None, result, fix_secure: bool, console: Console) -> None:
    """Apply safe (and optionally secure) fixes to the project in-place.

    Args:
        root: Lint target path (app dir or app-store dir)
        app_id: Optional specific app to fix
        result: LintingResult containing errors
        fix_secure: Whether to apply secure fixes (e.g., add user: "1000:1000")
        console: Rich console for status output
    """
    # Group errors by file path to minimize IO
    by_file: dict[str, list[LintingError]] = {}
    for err in result.errors:
        by_file.setdefault(err.file, []).append(err)

    is_single_app_root = (root / "umbrel-app.yml").exists()

    for file_rel, errs in by_file.items():
        # Normalize file path relative to the provided root
        rel_path = file_rel
        # When linting a single app folder, reported files are prefixed with
        # the app_id (usually equal to root.name). Strip that prefix.
        if is_single_app_root:
            prefix = f"{(app_id or root.name)}/"
            if file_rel.startswith(prefix):
                rel_path = file_rel[len(prefix) :]

        file_path = (root / rel_path).resolve()
        if not str(file_path).startswith(str(root.resolve())):
            continue  # safety guard

        # Fix directory structure (empty_app_data_directory)
        for err in errs:
            if err.id == "empty_app_data_directory":
                dir_path = (root / err.file).resolve()
                gitkeep = dir_path / ".gitkeep"
                if write_file_safely(gitkeep, ""):
                    console.print(f"[green]Created[/green] {gitkeep}")

        # Fix missing APP_DATA_DIR mounted directories (skip files)
        for err in errs:
            if err.id == "missing_file_or_directory":
                # Only act on directories (best-effort heuristic)
                # We avoid creating known file targets like *.conf by default
                missing = err.title.split('"')[1] if '"' in err.title else None
                if not missing:
                    continue
                # Derive absolute path under root
                abs_missing = (root / missing.lstrip("/")).resolve()
                # Skip if looks like a file (has an extension)
                if abs_missing.suffix:
                    continue
                abs_missing.mkdir(parents=True, exist_ok=True)
                write_file_safely(abs_missing / ".gitkeep", "")
                console.print(f"[green]Ensured directory[/green] {abs_missing} (+ .gitkeep)")

        # YAML-based fixes for docker-compose.yml and umbrel-app.yml
        if file_path.name in ("docker-compose.yml", "docker-compose.yaml", "umbrel-app.yml"):
            try:
                content = file_path.read_text(encoding="utf-8")
                data = yaml.safe_load(content) or {}
            except Exception:
                continue

            changed = False

            if file_path.name.startswith("docker-compose") and isinstance(data, dict):
                services = data.get("services", {})
                # invalid_yaml_boolean_value -> quote booleans
                for svc_name, svc in services.items():
                    if not isinstance(svc, dict):
                        continue
                    for key in ("environment", "labels", "extra_hosts"):
                        kv = svc.get(key)
                        if isinstance(kv, dict):
                            for k, v in list(kv.items()):
                                if isinstance(v, bool):
                                    kv[k] = str(v).lower()
                                    changed = True
                    # invalid_restart_policy -> ensure on-failure (skip app_proxy)
                    if svc_name != "app_proxy":
                        if svc.get("restart") != "on-failure":
                            svc["restart"] = "on-failure"
                            changed = True
                    # invalid_container_user -> secure fix
                    if fix_secure:
                        # Never enforce user for the special app_proxy service
                        if svc_name == "app_proxy":
                            # If we previously added user: "1000:1000" by mistake, remove it
                            if svc.get("user") == "1000:1000":
                                del svc["user"]
                                changed = True
                            # Skip further user handling for app_proxy
                            continue
                        user = svc.get("user")
                        env = svc.get("environment")
                        has_uid = False
                        if isinstance(env, dict):
                            has_uid = str(env.get("UID")) == "1000" or str(env.get("PUID")) == "1000"
                        elif isinstance(env, list):
                            has_uid = any(e in ("UID=1000", "PUID=1000") for e in env)
                        # Set a non-root user if missing and no UID/PUID explicitly set
                        if (user is None or user == "root") and not has_uid:
                            svc["user"] = "1000:1000"
                            changed = True

            if file_path.name == "umbrel-app.yml":
                # Perform a minimal, line-based edit to only change the tagline line,
                # preserving the rest of the file formatting (e.g., description block styles).
                lines = content.splitlines(keepends=True)
                text_changed = False
                for i, line in enumerate(lines):
                    stripped = line.lstrip()
                    if not stripped.startswith("tagline:"):
                        continue
                    # Preserve indentation and trailing whitespace
                    indent_len = len(line) - len(stripped)
                    prefix = line[:indent_len]  # indentation
                    # Split on first ':'
                    before, _, after = stripped.partition(":")
                    value = after.strip()
                    if not value:
                        continue
                    # Preserve surrounding quotes if present
                    quote = None
                    if (value.startswith('"') and value.endswith('"') and len(value) >= 2) or (
                        value.startswith("'") and value.endswith("'") and len(value) >= 2
                    ):
                        quote = value[0]
                        inner = value[1:-1]
                    else:
                        inner = value
                    # Apply rule: remove trailing period only if exactly one period and at end
                    if inner.endswith(".") and inner.count(".") == 1:
                        inner = inner[:-1]
                        # Reconstruct line
                        new_value = f"{quote}{inner}{quote}" if quote else inner
                        lines[i] = f"{prefix}tagline: {new_value}\n"
                        text_changed = True
                    # Only modify the first tagline occurrence
                    break
                if text_changed:
                    file_path.write_text("".join(lines), encoding="utf-8")
                    console.print(f"[green]Updated[/green] {file_path}")
            else:
                if changed:
                    file_path.write_text(yaml.safe_dump(data, sort_keys=False), encoding="utf-8")
                    console.print(f"[green]Updated[/green] {file_path}")





"""
Main linter implementation.

This module provides the core UmbrelLinter class that orchestrates
all linting operations for Umbrel applications and app stores.
"""

from __future__ import annotations

from pathlib import Path

from ..schemas.umbrel_app import UmbrelAppManifest, UmbrelAppStoreManifest
from ..utils.filesystem import file_exists, get_directory_files
from ..validators.docker_compose_validator import DockerComposeValidator
from ..validators.github_validator import GitHubValidator
from ..validators.yaml_validator import parse_yaml_with_error_handling
from .models import (
    FileEntry,
    LinterConfig,
    LintingContext,
    LintingError,
    LintingResult,
    Severity,
)


class UmbrelLinter:
    """
    Main linter class for Umbrel applications and app stores.
    
    This class provides comprehensive linting capabilities including:
    - YAML validation for app manifests
    - Docker Compose validation
    - Directory structure validation
    - Security and best practices checks
    """

    def __init__(self, config: LinterConfig | None = None):
        """
        Initialize the linter.
        
        Args:
            config: Optional linter configuration
        """
        self.config = config or LinterConfig()
        self.docker_compose_validator = DockerComposeValidator()
        self.github_validator = GitHubValidator()

    def lint_app_store(self, directory: Path, context: LintingContext | None = None) -> LintingResult:
        """
        Lint an Umbrel app store.
        
        Args:
            directory: Path to the app store directory
            context: Optional linting context
            
        Returns:
            Linting result with all errors found
        """
        result = LintingResult()
        context = context or LintingContext()

        # Check for umbrel-app-store.yml
        store_manifest_path = directory / "umbrel-app-store.yml"
        if not file_exists(store_manifest_path):
            if context.app_store_type == "community":
                result.add_error(LintingError(
                    id="missing_umbrel_app_store_yml",
                    severity=Severity.ERROR,
                    title="umbrel-app-store.yml does not exist",
                    message="For community app stores, the file umbrel-app-store.yml is required",
                    file="umbrel-app-store.yml",
                ))
        else:
            # Validate umbrel-app-store.yml
            store_errors = self._validate_app_store_manifest(store_manifest_path)
            for error in store_errors:
                result.add_error(error)

        # Check for README.md
        readme_path = directory / "README.md"
        if not file_exists(readme_path):
            result.add_error(LintingError(
                id="missing_readme",
                severity=Severity.WARNING,
                title="README.md does not exist",
                message="A README.md file is highly recommended to tell users, how to install your App Store and what apps are available",
                file="README.md",
            ))

        return result

    async def lint_app(self, directory: Path, app_id: str, context: LintingContext | None = None) -> LintingResult:
        """
        Lint a single Umbrel app.
        
        Args:
            directory: Path to the app directory
            app_id: ID of the app to lint
            context: Optional linting context
            
        Returns:
            Linting result with all errors found
        """
        result = LintingResult()
        context = context or LintingContext()
        context.app_id = app_id

        app_directory = directory / app_id
        if not app_directory.exists():
            result.add_error(LintingError(
                id="app_directory_not_found",
                severity=Severity.ERROR,
                title=f"App directory not found: {app_id}",
                message=f"App with id {app_id} does not exist",
                file=app_id,
            ))
            return result

        # Get all files in the app directory
        files = get_directory_files(app_directory)

        # Validate umbrel-app.yml
        manifest_errors = await self._validate_app_manifest(app_directory / "umbrel-app.yml", app_id, context)
        for error in manifest_errors:
            result.add_error(error)

        # Validate docker-compose.yml
        compose_errors = await self._validate_docker_compose(
            app_directory / "docker-compose.yml",
            app_id,
            files,
            context,
        )
        for error in compose_errors:
            result.add_error(error)

        # Validate directory structure
        structure_errors = self._validate_directory_structure(files)
        for error in structure_errors:
            result.add_error(error)

        return result

    async def lint_all_apps(self, directory: Path, context: LintingContext | None = None) -> LintingResult:
        """
        Lint all apps in a directory.
        
        Args:
            directory: Path to the directory containing apps
            context: Optional linting context
            
        Returns:
            Combined linting result for all apps
        """
        result = LintingResult()
        context = context or LintingContext()

        # Check if this directory itself is an app (contains umbrel-app.yml)
        if (directory / "umbrel-app.yml").exists():
            # This is a single app directory, not an app store
            # Get the app ID from the directory name
            app_id = directory.name
            app_result = await self.lint_app(directory.parent, app_id, context)
            return app_result

        # First lint the app store itself
        store_result = self.lint_app_store(directory, context)
        result.errors.extend(store_result.errors)
        result.success = result.success and store_result.success

        # Find all app directories
        app_directories = []
        for item in directory.iterdir():
            if item.is_dir() and not item.name.startswith("."):
                # Check if it's an app directory by looking for umbrel-app.yml
                if (item / "umbrel-app.yml").exists():
                    app_directories.append(item.name)

        # Lint each app
        for app_id in app_directories:
            app_result = await self.lint_app(directory, app_id, context)
            result.errors.extend(app_result.errors)
            result.success = result.success and app_result.success

        return result

    def _validate_app_store_manifest(self, manifest_path: Path) -> list[LintingError]:
        """Validate umbrel-app-store.yml manifest."""
        errors = []

        try:
            with open(manifest_path, encoding="utf-8") as f:
                content = f.read()
        except OSError as e:
            errors.append(LintingError(
                id="file_read_error",
                severity=Severity.ERROR,
                title="Failed to read umbrel-app-store.yml",
                message=str(e),
                file=str(manifest_path),
            ))
            return errors

        # Parse YAML
        data, parse_error = parse_yaml_with_error_handling(content, "umbrel-app-store.yml")
        if parse_error:
            errors.append(parse_error)
            return errors

        # Validate with Pydantic schema
        try:
            UmbrelAppStoreManifest(**data)
        except Exception as e:
            errors.append(LintingError(
                id="schema_validation_error",
                severity=Severity.ERROR,
                title="umbrel-app-store.yml validation failed",
                message=str(e),
                file="umbrel-app-store.yml",
            ))

        return errors

    async def _validate_app_manifest(self, manifest_path: Path, app_id: str, context: LintingContext) -> list[LintingError]:
        """Validate umbrel-app.yml manifest."""
        errors = []

        if not file_exists(manifest_path):
            errors.append(LintingError(
                id="missing_umbrel_app_yml",
                severity=Severity.ERROR,
                title="umbrel-app.yml does not exist",
                message='Every app needs a manifest file called "umbrel-app.yml" at the root of the app directory',
                file=f"{app_id}/umbrel-app.yml",
            ))
            return errors

        try:
            with open(manifest_path, encoding="utf-8") as f:
                content = f.read()
        except OSError as e:
            errors.append(LintingError(
                id="file_read_error",
                severity=Severity.ERROR,
                title="Failed to read umbrel-app.yml",
                message=str(e),
                file=f"{app_id}/umbrel-app.yml",
            ))
            return errors

        # Parse YAML
        data, parse_error = parse_yaml_with_error_handling(content, f"{app_id}/umbrel-app.yml")
        if parse_error:
            errors.append(parse_error)
            return errors

        # Validate GitHub URLs (submission and repo) first
        github_errors = await self.github_validator.validate_github_urls(data, app_id)
        errors.extend(github_errors)

        # Validate with Pydantic schema
        try:
            manifest = UmbrelAppManifest(**data)
        except Exception as e:
            # Parse Pydantic validation errors for better formatting
            error_message = str(e)
            
            # Try to extract field name and error from Pydantic error
            if "validation error" in error_message.lower():
                lines = error_message.split('\n')
                field_name = None
                error_detail = None
                
                # Parse the error message to extract field and error
                for i, line in enumerate(lines):
                    line = line.strip()
                    # Skip header lines
                    if "validation error" in line.lower() or not line:
                        continue
                    # Look for field name (lines without "For further information")
                    if not line.startswith("For further information") and not any(keyword in line.lower() for keyword in ["input should", "field required", "value error"]):
                        if not field_name:
                            field_name = line
                    # Look for error detail
                    elif any(keyword in line.lower() for keyword in ["input should", "field required", "value error"]):
                        error_detail = line
                        break
                
                # Create user-friendly error message
                if field_name and error_detail:
                    # Handle specific error types
                    if "tagline" in error_message.lower() and "period" in error_message.lower():
                        title = "Invalid tagline"
                        message = "Taglines should not end with a period"
                    elif "input should be a valid string" in error_detail.lower():
                        title = f'Invalid field "{field_name}"'
                        message = f'The field "{field_name}" must be a string, but received an invalid value'
                    elif "field required" in error_detail.lower():
                        title = f'Missing required field "{field_name}"'
                        message = f'The required field "{field_name}" is missing'
                    else:
                        title = f'Invalid field "{field_name}"'
                        message = error_detail
                    
                    errors.append(LintingError(
                        id="schema_validation_error",
                        severity=Severity.ERROR,
                        title=title,
                        message=message,
                        file=f"{app_id}/umbrel-app.yml",
                        properties_path=field_name if field_name else None,
                    ))
                else:
                    # Fallback to generic error
                    errors.append(LintingError(
                        id="schema_validation_error",
                        severity=Severity.ERROR,
                        title="umbrel-app.yml validation failed",
                        message=error_message,
                        file=f"{app_id}/umbrel-app.yml",
                    ))
            else:
                # Non-Pydantic error
                errors.append(LintingError(
                    id="schema_validation_error",
                    severity=Severity.ERROR,
                    title="umbrel-app.yml validation failed",
                    message=error_message,
                    file=f"{app_id}/umbrel-app.yml",
                ))
            # Don't return early - continue with other validations
            manifest = None

        # Additional validations for new submissions (only if manifest is valid)
        if context.is_new_submission and manifest is not None:
            # Check submission field matches PR URL
            if context.pull_request_url and manifest.submission != context.pull_request_url:
                errors.append(LintingError(
                    id="invalid_submission_field",
                    severity=Severity.ERROR,
                    title=f'Invalid submission field "{manifest.submission}"',
                    message=f"The submission field must be set to the URL of this pull request: {context.pull_request_url}",
                    file=f"{app_id}/umbrel-app.yml",
                    properties_path="submission",
                ))

            # Check release notes are empty for new submissions
            if manifest.releaseNotes and len(manifest.releaseNotes) > 0:
                errors.append(LintingError(
                    id="filled_out_release_notes_on_first_submission",
                    severity=Severity.ERROR,
                    title='"releaseNotes" needs to be empty for new app submissions',
                    message='The "releaseNotes" field must be empty for new app submissions as it is being displayed to the user only in case of an update.',
                    file=f"{app_id}/umbrel-app.yml",
                    properties_path="releaseNotes",
                ))

            # Check icon and gallery are empty for new submissions
            if manifest.icon or (manifest.gallery and len(manifest.gallery) > 0):
                errors.append(LintingError(
                    id="filled_out_icon_or_gallery_on_first_submission",
                    severity=Severity.WARNING,
                    title='"icon" and "gallery" needs to be empty for new app submissions',
                    message='The "icon" and "gallery" fields must be empty for new app submissions as it is being created by the Umbrel team.',
                    file=f"{app_id}/umbrel-app.yml",
                    properties_path="icon" if manifest.icon else "gallery",
                ))

        # Check for duplicate ports
        if context.all_app_manifests:
            used_ports = self._get_used_ports(context.all_app_manifests, app_id)
            if manifest.port in used_ports:
                app_name = used_ports[manifest.port]
                errors.append(LintingError(
                    id="duplicate_ui_port",
                    severity=Severity.ERROR,
                    title=f"Port {manifest.port} is already used by {app_name}",
                    message="Each app must use a unique port",
                    file=f"{app_id}/umbrel-app.yml",
                    properties_path="port",
                ))

        return errors

    async def _validate_docker_compose(self, compose_path: Path, app_id: str, files: list[FileEntry], context: LintingContext) -> list[LintingError]:
        """Validate docker-compose.yml file."""
        errors = []

        if not file_exists(compose_path):
            errors.append(LintingError(
                id="missing_docker_compose_yml",
                severity=Severity.ERROR,
                title="docker-compose.yml does not exist",
                message='Every app needs a docker compose file called "docker-compose.yml" at the root of the app directory',
                file=f"{app_id}/docker-compose.yml",
            ))
            return errors

        try:
            with open(compose_path, encoding="utf-8") as f:
                content = f.read()
        except OSError as e:
            errors.append(LintingError(
                id="file_read_error",
                severity=Severity.ERROR,
                title="Failed to read docker-compose.yml",
                message=str(e),
                file=f"{app_id}/docker-compose.yml",
            ))
            return errors

        # Use Docker Compose validator
        options = {
            "check_image_architectures": self.config.check_image_architectures,
        }

        compose_errors = await self.docker_compose_validator.validate_docker_compose(
            content, app_id, files, options,
        )
        errors.extend(compose_errors)

        return errors

    def _validate_directory_structure(self, files: list[FileEntry]) -> list[LintingError]:
        """Validate directory structure."""
        errors = []

        # Find empty directories (no .gitkeep file)
        directories = {f.path for f in files if f.type == "directory"}
        files_set = {f.path for f in files if f.type == "file"}

        for directory in directories:
            # Check if directory has any files or subdirectories
            has_content = any(
                f.path.startswith(directory + "/") for f in files
            )

            if not has_content:
                errors.append(LintingError(
                    id="empty_app_data_directory",
                    severity=Severity.ERROR,
                    title=f'Empty directory "{directory}"',
                    message=f'Please add a ".gitkeep" file to the directory "{directory}". This is necessary to ensure the correct permissions of the directory after cloning!',
                    file=directory,
                ))

        return errors

    def _get_used_ports(self, all_manifests: list[str], current_app_id: str) -> dict[int, str]:
        """Get used ports from all app manifests."""
        used_ports = {}

        for manifest_content in all_manifests:
            data, _ = parse_yaml_with_error_handling(manifest_content, "manifest")
            if data and isinstance(data, dict):
                app_id = data.get("id")
                name = data.get("name")
                port = data.get("port")

                if app_id and name and port and app_id != current_app_id:
                    # Ensure port is an integer
                    try:
                        port_int = int(port)
                        used_ports[port_int] = f"{name} ({app_id})"
                    except (ValueError, TypeError):
                        # Skip invalid port values
                        continue

        return used_ports



"""
File system utilities.

This module provides utility functions for file system operations
used by the linter.
"""

from __future__ import annotations

from pathlib import Path

from ..core.models import FileEntry


def file_exists(file_path: Path) -> bool:
    """
    Check if a file exists.
    
    Args:
        file_path: Path to the file
        
    Returns:
        True if the file exists, False otherwise
    """
    return file_path.exists() and file_path.is_file()


def directory_exists(dir_path: Path) -> bool:
    """
    Check if a directory exists.
    
    Args:
        dir_path: Path to the directory
        
    Returns:
        True if the directory exists, False otherwise
    """
    return dir_path.exists() and dir_path.is_dir()


def get_directory_files(directory: Path, relative_to: Path | None = None) -> list[FileEntry]:
    """
    Recursively get all files and directories in a directory.
    
    Args:
        directory: Directory to scan
        relative_to: Optional base path for relative paths
        
    Returns:
        List of file entries
    """
    entries = []
    relative_to = relative_to or directory

    try:
        for item in directory.rglob("*"):
            if item.is_file() or item.is_dir():
                # Get relative path
                try:
                    relative_path = item.relative_to(relative_to)
                except ValueError:
                    # If we can't get relative path, use absolute
                    relative_path = item

                entry = FileEntry(
                    path=str(relative_path),
                    type="file" if item.is_file() else "directory",
                )
                entries.append(entry)
    except (OSError, PermissionError):
        # Handle permission errors or other filesystem issues
        pass

    return entries


def read_file_safely(file_path: Path, encoding: str = "utf-8") -> str | None:
    """
    Safely read a file with error handling.
    
    Args:
        file_path: Path to the file
        encoding: File encoding
        
    Returns:
        File content as string, or None if reading failed
    """
    try:
        with open(file_path, encoding=encoding) as f:
            return f.read()
    except (OSError, UnicodeDecodeError):
        return None


def write_file_safely(file_path: Path, content: str, encoding: str = "utf-8") -> bool:
    """
    Safely write content to a file.
    
    Args:
        file_path: Path to the file
        content: Content to write
        encoding: File encoding
        
    Returns:
        True if successful, False otherwise
    """
    try:
        # Ensure parent directory exists
        file_path.parent.mkdir(parents=True, exist_ok=True)

        with open(file_path, "w", encoding=encoding) as f:
            f.write(content)
        return True
    except OSError:
        return False


def get_file_size(file_path: Path) -> int:
    """
    Get file size in bytes.
    
    Args:
        file_path: Path to the file
        
    Returns:
        File size in bytes, or 0 if file doesn't exist
    """
    try:
        return file_path.stat().st_size
    except OSError:
        return 0


def is_empty_directory(dir_path: Path) -> bool:
    """
    Check if a directory is empty.
    
    Args:
        dir_path: Path to the directory
        
    Returns:
        True if directory is empty, False otherwise
    """
    try:
        return not any(dir_path.iterdir())
    except (OSError, PermissionError):
        return False


def find_files_by_pattern(directory: Path, pattern: str) -> list[Path]:
    """
    Find files matching a pattern in a directory.
    
    Args:
        directory: Directory to search
        pattern: File pattern (e.g., "*.yml", "*.yaml")
        
    Returns:
        List of matching file paths
    """
    matches = []
    try:
        for file_path in directory.rglob(pattern):
            if file_path.is_file():
                matches.append(file_path)
    except (OSError, PermissionError):
        pass

    return matches


def get_relative_path(file_path: Path, base_path: Path) -> str:
    """
    Get relative path from base path.
    
    Args:
        file_path: File path
        base_path: Base path
        
    Returns:
        Relative path as string
    """
    try:
        return str(file_path.relative_to(base_path))
    except ValueError:
        return str(file_path)



"""
Schema definitions for umbrel-app.yml validation.

This module defines Pydantic models for validating Umbrel app manifest files,
ensuring they conform to the expected structure and requirements.
"""

from __future__ import annotations

from pydantic import BaseModel, Field, HttpUrl, validator


class UmbrelAppManifest(BaseModel):
    """Schema for umbrel-app.yml manifest files."""

    # Required fields
    manifestVersion: float | int = Field(
        ...,
        description="Manifest version (1, 1.1, or 1.2)",
        ge=1,
        le=2,
    )
    id: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Unique identifier for the app",
    )
    name: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Display name of the app",
    )
    tagline: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description="Short tagline for the app",
    )
    category: str = Field(
        ...,
        description="Category of the app",
    )
    version: str = Field(
        ...,
        min_length=1,
        description="Version of the app",
    )
    port: int = Field(
        ...,
        ge=0,
        le=65535,
        description="Port number for the app",
    )
    description: str = Field(
        ...,
        min_length=1,
        max_length=5000,
        description="Detailed description of the app",
    )
    developer: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Name of the developer",
    )
    submitter: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Name of the submitter",
    )
    
    @validator("developer", "submitter", pre=True)
    def convert_to_string(cls, v):
        """Convert numeric values to strings for developer and submitter fields."""
        if isinstance(v, (int, float)):
            return str(v)
        return v
    submission: HttpUrl = Field(
        ...,
        description="URL of the submission (usually a pull request)",
    )
    support: HttpUrl = Field(
        ...,
        description="Support URL for the app",
    )
    website: HttpUrl = Field(
        ...,
        description="Website URL for the app",
    )
    path: str = Field(
        ...,
        description="Path for the app (can be empty string)",
    )

    # Optional fields
    disabled: bool | None = Field(None, description="Whether the app is disabled")
    icon: str | None = Field(None, description="Icon URL for the app")
    gallery: list[str] = Field(default_factory=list, description="Gallery images")
    releaseNotes: str | None = Field(
        None,
        min_length=0,
        max_length=5000,
        description="Release notes for the app",
    )
    dependencies: list[str] = Field(
        default_factory=list,
        description="List of app dependencies",
    )
    permissions: list[str] | None = Field(
        None,
        description="Required permissions (STORAGE_DOWNLOADS, GPU)",
    )
    defaultUsername: str | None = Field(
        None,
        description="Default username for the app",
    )
    defaultPassword: str | None = Field(
        None,
        description="Default password for the app",
    )
    deterministicPassword: bool | None = Field(
        None,
        description="Whether to use deterministic password",
    )
    optimizedForUmbrelHome: bool | None = Field(
        None,
        description="Whether optimized for Umbrel Home",
    )
    torOnly: bool | None = Field(
        None,
        description="Whether the app is Tor-only",
    )
    installSize: int | None = Field(
        None,
        ge=0,
        description="Install size in bytes",
    )
    widgets: list[dict] | None = Field(
        None,
        description="Widget configurations",
    )
    defaultShell: str | None = Field(
        None,
        description="Default shell for the app",
    )
    backupIgnore: list[str] | None = Field(
        None,
        description="Files to ignore during backup",
    )
    repo: HttpUrl | str | None = Field(
        None,
        description="Repository URL (can be empty string)",
    )

    @validator("manifestVersion")
    def validate_manifest_version(cls, v):
        """Validate manifest version is supported."""
        if v not in [1, 1.1, 1.2]:
            raise ValueError("manifestVersion must be 1, 1.1, or 1.2")
        return v

    @validator("id")
    def validate_id(cls, v):
        """Validate app ID doesn't start with reserved prefix."""
        if v.startswith("umbrel-app-store"):
            raise ValueError(
                "The id of the app can't start with 'umbrel-app-store' as it is the id of the app repository",
            )
        return v

    @validator("tagline")
    def validate_tagline(cls, v):
        """Validate tagline doesn't end with period."""
        if v.endswith(".") and v.count(".") == 1:
            raise ValueError("Taglines should not end with a period")
        return v

    @validator("category")
    def validate_category(cls, v):
        """Validate category is from allowed list."""
        allowed_categories = [
            "files", "bitcoin", "media", "networking", "social",
            "automation", "finance", "ai", "developer",
        ]
        if v not in allowed_categories:
            raise ValueError(f'category must be one of: {", ".join(allowed_categories)}')
        return v

    @validator("permissions")
    def validate_permissions(cls, v):
        """Validate permissions are from allowed list."""
        if v is not None:
            allowed_permissions = ["STORAGE_DOWNLOADS", "GPU"]
            for perm in v:
                if perm not in allowed_permissions:
                    raise ValueError(f'permission must be one of: {", ".join(allowed_permissions)}')
        return v

    @validator("dependencies")
    def validate_dependencies(cls, v, values):
        """Validate dependencies don't include self."""
        if "id" in values and values["id"] in v:
            raise ValueError("Dependencies can't include its own app id")
        return v

    @validator("path")
    def validate_path(cls, v):
        """Validate path is a valid URL path (can be empty)."""
        # Path can be empty string according to TypeScript schema
        if v and not v.startswith("/"):
            raise ValueError("path must start with /")
        return v

    @validator("repo")
    def validate_repo(cls, v):
        """Validate repo is either empty string or valid URL."""
        if v == "":
            return v
        return v


class UmbrelAppStoreManifest(BaseModel):
    """Schema for umbrel-app-store.yml manifest files."""

    id: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Unique identifier for the app store",
    )
    name: str = Field(
        ...,
        min_length=1,
        max_length=50,
        description="Display name of the app store",
    )

    @validator("id")
    def validate_id(cls, v):
        """Validate app store ID format and restrictions."""
        if v.startswith("umbrel-app-store"):
            raise ValueError(
                "The id of the app can't start with 'umbrel-app-store' as it is the id of the official Umbrel App Store.",
            )

        # Check format: only lowercase letters and dashes
        import re
        if not re.match(r"^[a-z]+(?:-[a-z]+)*$", v):
            raise ValueError(
                "The id of the app should contain only alphabets ('a' to 'z') and dashes ('-').",
            )

        return v



"""
Docker Compose validation utilities.

This module provides comprehensive validation for Docker Compose files,
including schema validation, security checks, and best practices enforcement.
"""

from __future__ import annotations

import re
from typing import Any

import jsonschema
from jsonschema import Draft7Validator

from ..core.models import FileEntry, LintingError, Severity
from .yaml_validator import parse_yaml_with_error_handling
from .docker_image_validator import DockerImageValidator
from .variable_mocker import VariableMocker

# Docker Compose JSON Schema (simplified version)
DOCKER_COMPOSE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "version": {"type": "string"},
        "services": {
            "type": "object",
            "patternProperties": {
                "^[a-zA-Z0-9._-]+$": {
                    "type": "object",
                    "properties": {
                        "image": {"type": "string"},
                        "user": {"type": "string"},
                        "restart": {"type": "string"},
                        "ports": {
                            "type": "array",
                            "items": {
                                "oneOf": [
                                    {"type": "string"},
                                    {"type": "number"},
                                    {"type": "object"},
                                ],
                            },
                        },
                        "volumes": {
                            "type": "array",
                            "items": {
                                "oneOf": [
                                    {"type": "string"},
                                    {"type": "object"},
                                ],
                            },
                        },
                        "environment": {
                            "oneOf": [
                                {"type": "object"},
                                {"type": "array", "items": {"type": "string"}},
                            ],
                        },
                        "labels": {
                            "oneOf": [
                                {"type": "object"},
                                {"type": "array", "items": {"type": "string"}},
                            ],
                        },
                        "extra_hosts": {
                            "oneOf": [
                                {"type": "object"},
                                {"type": "array", "items": {"type": "string"}},
                            ],
                        },
                        "network_mode": {"type": "string"},
                        "hostname": {"type": "string"},
                        "container_name": {"type": "string"},
                    },
                },
            },
        },
    },
}


class DockerComposeValidator:
    """Validator for Docker Compose files."""

    def __init__(self):
        """Initialize the validator."""
        self.schema_validator = Draft7Validator(DOCKER_COMPOSE_SCHEMA)
        self.image_validator = DockerImageValidator()
        self.variable_mocker = VariableMocker()

    async def validate_docker_compose(
        self,
        content: str,
        app_id: str,
        files: list[FileEntry],
        options: dict[str, Any] | None = None,
    ) -> list[LintingError]:
        """
        Validate a Docker Compose file.
        
        Args:
            content: Docker Compose YAML content
            app_id: ID of the app being validated
            files: List of files in the app directory
            options: Validation options
            
        Returns:
            List of linting errors
        """
        errors = []
        options = options or {}

        # Mock variables in the content
        mocked_content = self.variable_mocker.mock_variables(content)
        
        # Parse YAML
        data, parse_error = parse_yaml_with_error_handling(content, f"{app_id}/docker-compose.yml")
        if parse_error:
            return [parse_error]
        
        # Parse mocked YAML for schema validation
        mocked_data, mocked_parse_error = parse_yaml_with_error_handling(mocked_content, f"{app_id}/docker-compose.yml")
        if mocked_parse_error:
            return [mocked_parse_error]

        # Validate against JSON Schema using mocked data
        schema_errors = self._validate_schema(mocked_data, app_id)
        errors.extend(schema_errors)

        # Validate image names and architectures
        services = data.get("services", {})
        check_architectures = options.get("check_image_architectures", False)
        image_errors = await self.image_validator.validate_images(services, app_id, check_architectures)
        errors.extend(image_errors)

        # Validate boolean values using mocked data
        boolean_errors = self._validate_boolean_values(mocked_data, app_id)
        errors.extend(boolean_errors)

        # Validate volume mounts using original data (for file path checking)
        volume_errors = self._validate_volume_mounts(data, app_id, files)
        errors.extend(volume_errors)

        # Validate security settings using mocked data
        security_errors = self._validate_security_settings(mocked_data, app_id)
        errors.extend(security_errors)

        # Validate port mappings using mocked data
        port_errors = self._validate_port_mappings(mocked_data, app_id)
        errors.extend(port_errors)

        # Validate app proxy configuration using mocked data
        proxy_errors = self._validate_app_proxy_configuration(mocked_data, app_id)
        errors.extend(proxy_errors)

        # Validate restart policies using mocked data
        restart_errors = self._validate_restart_policies(mocked_data, app_id)
        errors.extend(restart_errors)

        return errors

    def _validate_schema(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate against Docker Compose JSON Schema."""
        errors = []

        try:
            self.schema_validator.validate(data)
        except jsonschema.ValidationError as e:
            error = LintingError(
                id="schema_validation_error",
                severity=Severity.ERROR,
                title="Docker Compose schema validation failed",
                message=str(e.message),
                file=f"{app_id}/docker-compose.yml",
                properties_path=e.json_path,
            )
            errors.append(error)

        return errors

    def _validate_image_names(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate Docker image names follow naming conventions."""
        errors = []
        services = data.get("services", {})

        for service_name, service_config in services.items():
            image = service_config.get("image")
            if not image:
                continue

            # Check image format: name:tag@digest
            image_match = re.match(r"^(.+):(.+)@(.+)$", image)
            if not image_match:
                errors.append(LintingError(
                    id="invalid_docker_image_name",
                    severity=Severity.ERROR,
                    title=f'Invalid image name "{image}"',
                    message='Images should be named like "<name>:<version-tag>@<sha256>"',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.image",
                ))
            else:
                _, tag, _ = image_match.groups()
                if tag == "latest":
                    errors.append(LintingError(
                        id="invalid_docker_image_name",
                        severity=Severity.WARNING,
                        title=f'Invalid image tag "{tag}"',
                        message='Images should not use the "latest" tag',
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.image",
                    ))

        return errors

    def _validate_boolean_values(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate that boolean values are strings for Docker Compose V1 compatibility."""
        errors = []
        services = data.get("services", {})

        for service_name, service_config in services.items():
            # Check environment variables
            env = service_config.get("environment")
            if isinstance(env, dict):
                for key, value in env.items():
                    if isinstance(value, bool):
                        errors.append(LintingError(
                            id="invalid_yaml_boolean_value",
                            severity=Severity.ERROR,
                            title=f'Invalid YAML boolean value for key "{key}"',
                            message=f'Boolean values should be strings like "{str(value).lower()}" instead of {value}',
                            file=f"{app_id}/docker-compose.yml",
                            properties_path=f"services.{service_name}.environment.{key}",
                        ))

            # Check labels
            labels = service_config.get("labels")
            if isinstance(labels, dict):
                for key, value in labels.items():
                    if isinstance(value, bool):
                        errors.append(LintingError(
                            id="invalid_yaml_boolean_value",
                            severity=Severity.ERROR,
                            title=f'Invalid YAML boolean value for key "{key}"',
                            message=f'Boolean values should be strings like "{str(value).lower()}" instead of {value}',
                            file=f"{app_id}/docker-compose.yml",
                            properties_path=f"services.{service_name}.labels.{key}",
                        ))

            # Check extra_hosts
            extra_hosts = service_config.get("extra_hosts")
            if isinstance(extra_hosts, dict):
                for key, value in extra_hosts.items():
                    if isinstance(value, bool):
                        errors.append(LintingError(
                            id="invalid_yaml_boolean_value",
                            severity=Severity.ERROR,
                            title=f'Invalid YAML boolean value for key "{key}"',
                            message=f'Boolean values should be strings like "{str(value).lower()}" instead of {value}',
                            file=f"{app_id}/docker-compose.yml",
                            properties_path=f"services.{service_name}.extra_hosts.{key}",
                        ))

        return errors

    def _validate_volume_mounts(self, data: dict[str, Any], app_id: str, files: list[FileEntry]) -> list[LintingError]:
        """Validate volume mounts."""
        errors = []
        services = data.get("services", {})
        file_paths = {f.path for f in files}

        for service_name, service_config in services.items():
            volumes = service_config.get("volumes", [])

            for volume in volumes:
                if isinstance(volume, str):
                    # Check for direct APP_DATA_DIR mounting
                    if re.search(r"\$\{?APP_DATA_DIR\}?/?:", volume):
                        errors.append(LintingError(
                            id="invalid_app_data_dir_volume_mount",
                            severity=Severity.WARNING,
                            title=f'Volume "{volume}"',
                            message='Volumes should not be mounted directly into the "${APP_DATA_DIR}" directory! Please use a subdirectory like "${APP_DATA_DIR}/data" instead.',
                            file=f"{app_id}/docker-compose.yml",
                            properties_path=f"services.{service_name}.volumes",
                        ))

                    # Check for missing files/directories
                    if re.search(r"\$\{?APP_DATA_DIR\}?", volume):
                        match = re.search(r"\$\{?APP_DATA_DIR\}?/?(.*?):", volume)
                        if match:
                            relative_path = match.group(1).strip()
                            candidate_paths = {relative_path, f"{app_id}/{relative_path}"}
                            if relative_path and not (candidate_paths & file_paths):
                                errors.append(LintingError(
                                    id="missing_file_or_directory",
                                    severity=Severity.INFO,
                                    title=f'Mounted file/directory "/{app_id}/{relative_path}" doesn\'t exist',
                                    message=f'The volume "{volume}" tries to mount the file/directory "/{app_id}/{relative_path}", but it is not present. This can lead to permission errors!',
                                    file=f"{app_id}/docker-compose.yml",
                                    properties_path=f"services.{service_name}.volumes",
                                ))

                elif isinstance(volume, dict):
                    source = volume.get("source", "")
                    target = volume.get("target", "")

                    # Check for direct APP_DATA_DIR mounting
                    if re.search(r"\$\{?APP_DATA_DIR\}?/?$", source):
                        errors.append(LintingError(
                            id="invalid_app_data_dir_volume_mount",
                            severity=Severity.WARNING,
                            title=f'Volume "{source}:{target}"',
                            message='Volumes should not be mounted directly into the "${APP_DATA_DIR}" directory! Please use a subdirectory like "source: ${APP_DATA_DIR}/data" and "target: /some/dir" instead.',
                            file=f"{app_id}/docker-compose.yml",
                            properties_path=f"services.{service_name}.volumes",
                        ))

                    # Check for missing files/directories
                    if re.search(r"\$\{?APP_DATA_DIR\}?", source):
                        match = re.search(r"\$\{?APP_DATA_DIR\}?/?(.*?)$", source)
                        if match:
                            relative_path = match.group(1).strip()
                            candidate_paths = {relative_path, f"{app_id}/{relative_path}"}
                            if relative_path and not (candidate_paths & file_paths):
                                errors.append(LintingError(
                                    id="missing_file_or_directory",
                                    severity=Severity.INFO,
                                    title=f'Mounted file/directory "/{app_id}/{relative_path}" doesn\'t exist',
                                    message=f'The volume "{source}:{target}" tries to mount the file/directory "/{app_id}/{relative_path}", but it is not present. This can lead to permission errors!',
                                    file=f"{app_id}/docker-compose.yml",
                                    properties_path=f"services.{service_name}.volumes",
                                ))

        return errors

    def _validate_security_settings(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate security-related settings."""
        errors = []
        services = data.get("services", {})

        for service_name, service_config in services.items():
            # Check for Docker socket mounting
            volumes = service_config.get("volumes", [])
            for volume in volumes:
                if isinstance(volume, str) and "/var/run/docker.sock" in volume:
                    errors.append(LintingError(
                        id="docker_socket_mount",
                        severity=Severity.WARNING,
                        title=f'Docker socket is mounted in "{service_name}"',
                        message=f'The volume "{volume}" mounts the Docker socket, which can be a security risk. Consider using docker-in-docker instead (see portainer as an example).',
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.volumes",
                    ))
                elif isinstance(volume, dict) and "/var/run/docker.sock" in volume.get("source", ""):
                    errors.append(LintingError(
                        id="docker_socket_mount",
                        severity=Severity.WARNING,
                        title=f'Docker socket is mounted in "{service_name}"',
                        message=f'The volume "{volume.get("source")}:{volume.get("target")}" mounts the Docker socket, which can be a security risk. Consider using docker-in-docker instead (see portainer as an example).',
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.volumes",
                    ))

            # Check for root user usage
            user = service_config.get("user")
            environment = service_config.get("environment", {})

            # Check for UID/PUID environment variables
            has_uid_env = False
            if isinstance(environment, dict):
                for key, value in environment.items():
                    if key in ["UID", "PUID"] and str(value) == "1000":
                        has_uid_env = True
                        break
            elif isinstance(environment, list):
                for env_var in environment:
                    if "=" in env_var:
                        key, value = env_var.split("=", 1)
                        if key in ["UID", "PUID"] and value == "1000":
                            has_uid_env = True
                            break

            # Skip user validation for app_proxy service as it needs root privileges
            if service_name == "app_proxy":
                continue

            if user == "root":
                errors.append(LintingError(
                    id="invalid_container_user",
                    severity=Severity.INFO,
                    title=f'Using unsafe user "{user}" in service "{service_name}"',
                    message=f'The user "{user}" can lead to security vulnerabilities. If possible please use a non-root user instead.',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.user",
                ))
            elif not user and not has_uid_env:
                errors.append(LintingError(
                    id="invalid_container_user",
                    severity=Severity.INFO,
                    title=f'Potentially using unsafe user in service "{service_name}"',
                    message='The default container user "root" can lead to security vulnerabilities. If you are using the root user, please try to specify a different user (e.g. "1000:1000") in the compose file or try to set the UID/PUID and GID/PGID environment variables to 1000.',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.user",
                ))

            # Check for host network mode
            network_mode = service_config.get("network_mode")
            if network_mode == "host":
                errors.append(LintingError(
                    id="container_network_mode_host",
                    severity=Severity.INFO,
                    title=f'Service "{service_name}" uses host network mode',
                    message="The host network mode can lead to security vulnerabilities. If possible please use the default bridge network mode and expose the necessary ports.",
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.network_mode",
                ))

        return errors

    def _validate_port_mappings(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate port mappings."""
        errors = []
        services = data.get("services", {})

        for service_name, service_config in services.items():
            ports = service_config.get("ports", [])

            for port in ports:
                if isinstance(port, (str, int)):
                    errors.append(LintingError(
                        id="external_port_mapping",
                        severity=Severity.INFO,
                        title=f'External port mapping "{port}"',
                        message="Port mappings may be unnecessary for the app to function correctly. Docker's internal DNS resolves container names to IP addresses within the same network. External access to the web interface is handled by the app_proxy container. Port mappings are only needed if external access is required to a port not proxied by the app_proxy, or if an app needs to expose multiple ports for its functionality (e.g., DHCP, DNS, P2P, etc.).",
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.ports",
                    ))
                elif isinstance(port, dict):
                    target = port.get("target")
                    published = port.get("published")
                    port_str = f"{target}{f':{published}' if published else ''}"
                    errors.append(LintingError(
                        id="external_port_mapping",
                        severity=Severity.INFO,
                        title=f'External port mapping "{port_str}"',
                        message="Port mappings may be unnecessary for the app to function correctly. Docker's internal DNS resolves container names to IP addresses within the same network. External access to the web interface is handled by the app_proxy container. Port mappings are only needed if external access is required to a port not proxied by the app_proxy, or if an app needs to expose multiple ports for its functionality (e.g., DHCP, DNS, P2P, etc.).",
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.ports",
                    ))

        return errors

    def _validate_app_proxy_configuration(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate app_proxy configuration."""
        errors = []
        services = data.get("services", {})

        # Collect hostnames from all services
        hostnames = set()
        for service_name, service_config in services.items():
            if service_name == "app_proxy":
                continue

            hostname = service_config.get("hostname")
            if hostname:
                hostnames.add(hostname)

            container_name = service_config.get("container_name")
            if container_name:
                hostnames.add(container_name)

        # Validate app_proxy service
        app_proxy = services.get("app_proxy")
        if app_proxy:
            environment = app_proxy.get("environment", {})
            env_vars = {}

            if isinstance(environment, dict):
                env_vars = environment
            elif isinstance(environment, list):
                for env_var in environment:
                    if "=" in env_var:
                        key, value = env_var.split("=", 1)
                        env_vars[key] = value

            # Check APP_HOST
            if "APP_HOST" not in env_vars:
                errors.append(LintingError(
                    id="invalid_app_proxy_configuration",
                    severity=Severity.ERROR,
                    title="Missing APP_HOST environment variable",
                    message='The app_proxy container needs to have the APP_HOST environment variable set to the hostname of the app_proxy container (e.g. "<app-id>_<web-container-name>_1").',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path="services.app_proxy.environment",
                ))
            else:
                app_host = str(env_vars["APP_HOST"])
                if not app_host.startswith("$") and app_host not in hostnames:
                    parts = app_host.split("_")
                    if len(parts) >= 3:
                        appid, container_name, number = parts[0], parts[1], parts[2]
                        if (appid != app_id.lower() or
                            container_name not in [s for s in services.keys() if s != "app_proxy"] or
                            number != "1"):
                            errors.append(LintingError(
                                id="invalid_app_proxy_configuration",
                                severity=Severity.WARNING,
                                title="Invalid APP_HOST environment variable",
                                message='The APP_HOST environment variable must be set to the hostname of the app_proxy container (e.g. "<app-id>_<web-container-name>_1").',
                                file=f"{app_id}/docker-compose.yml",
                                properties_path="services.app_proxy.environment",
                            ))

            # Check APP_PORT
            if "APP_PORT" not in env_vars:
                errors.append(LintingError(
                    id="invalid_app_proxy_configuration",
                    severity=Severity.ERROR,
                    title="Missing APP_PORT environment variable",
                    message="The app_proxy container needs to have the APP_PORT environment variable set to the port the ui of the app inside the container is listening on.",
                    file=f"{app_id}/docker-compose.yml",
                    properties_path="services.app_proxy.environment",
                ))
            else:
                app_port = str(env_vars["APP_PORT"])
                if not app_port.startswith("$") and not app_port.isdigit():
                    errors.append(LintingError(
                        id="invalid_app_proxy_configuration",
                        severity=Severity.WARNING,
                        title="Invalid APP_PORT environment variable",
                        message="The APP_PORT environment variable must be set to the port the ui of the app inside the container is listening on.",
                        file=f"{app_id}/docker-compose.yml",
                        properties_path="services.app_proxy.environment",
                    ))

        return errors

    def _validate_restart_policies(self, data: dict[str, Any], app_id: str) -> list[LintingError]:
        """Validate restart policies."""
        errors = []
        services = data.get("services", {})

        for service_name, service_config in services.items():
            if service_name == "app_proxy":
                continue

            restart = service_config.get("restart")
            # Warn when restart is missing or not set to on-failure
            if restart != "on-failure":
                errors.append(LintingError(
                    id="invalid_restart_policy",
                    severity=Severity.WARNING,
                    title="Invalid restart policy",
                    message=f'The restart policy of the container "{service_name}" should be set to "on-failure".',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.restart",
                ))

        return errors



"""
GitHub validation utilities.

This module provides functionality to validate GitHub pull requests
and other GitHub-related resources.
"""

from __future__ import annotations

import re
from typing import Any
import httpx
from urllib.parse import urlparse

from ..core.models import LintingError, Severity


class GitHubValidator:
    """Validator for GitHub resources."""
    
    def __init__(self, timeout: float = 5.0):
        """Initialize the GitHub validator."""
        self.timeout = timeout
        self._cache: dict[str, bool] = {}
    
    async def validate_pull_request(self, pr_url: str, app_id: str) -> list[LintingError]:
        """
        Validate that a GitHub pull request URL exists and is accessible.
        
        Args:
            pr_url: GitHub pull request URL
            app_id: ID of the app being validated
            
        Returns:
            List of linting errors
        """
        errors = []
        
        # Check if URL is a valid GitHub PR URL
        if not self._is_github_pr_url(pr_url):
            errors.append(LintingError(
                id="invalid_submission_field",
                severity=Severity.ERROR,
                title=f'Invalid submission URL "{pr_url}"',
                message="Submission URL must be a valid GitHub pull request URL (e.g., https://github.com/owner/repo/pull/123)",
                file=f"{app_id}/umbrel-app.yml",
                properties_path="submission",
            ))
            return errors
        
        # Check if PR exists and is accessible
        if not await self._pr_exists(pr_url):
            errors.append(LintingError(
                id="invalid_submission_field",
                severity=Severity.ERROR,
                title=f'Pull request not found "{pr_url}"',
                message="The specified pull request does not exist or is not accessible. Please check the URL and ensure the PR is public.",
                file=f"{app_id}/umbrel-app.yml",
                properties_path="submission",
            ))
        
        return errors
    
    def _is_github_pr_url(self, url: str) -> bool:
        """Check if URL is a valid GitHub pull request URL."""
        try:
            parsed = urlparse(url)
            
            # Must be GitHub domain
            if parsed.hostname not in ["github.com", "www.github.com"]:
                return False
            
            # Must be HTTPS
            if parsed.scheme != "https":
                return False
            
            # Must match GitHub PR path pattern: /owner/repo/pull/number
            path_pattern = r"^/[^/]+/[^/]+/pull/\d+/?$"
            return bool(re.match(path_pattern, parsed.path))
            
        except Exception:
            return False
    
    async def _pr_exists(self, pr_url: str) -> bool:
        """Check if a GitHub pull request exists and is accessible.

        Uses GitHub REST API if possible for reliable status; falls back to HTML.
        """
        if pr_url in self._cache:
            return self._cache[pr_url]

        try:
            async with httpx.AsyncClient(timeout=self.timeout, follow_redirects=True) as client:
                headers = {
                    "User-Agent": "umbrel-linter/1.0",
                    "Accept": "application/vnd.github.v3+json",
                }

                owner, repo, number = self._parse_github_pr(pr_url)
                exists = True
                if owner and repo and number:
                    api_url = f"https://api.github.com/repos/{owner}/{repo}/pulls/{number}"
                    resp = await client.get(api_url, headers=headers)
                    if resp.status_code == 200:
                        exists = True
                    elif resp.status_code == 404:
                        exists = False
                    else:
                        # On rate-limit or other statuses, do not block lint; assume exists
                        exists = True
                else:
                    # Fallback to fetching the HTML PR page
                    resp = await client.get(pr_url, headers={"User-Agent": "umbrel-linter/1.0"})
                    if resp.status_code == 200:
                        exists = True
                    elif resp.status_code == 404:
                        exists = False
                    else:
                        exists = True

                self._cache[pr_url] = exists
                return exists
        except Exception:
            # On any error, do not fail lint for existence check
            self._cache[pr_url] = True
            return True

    def _parse_github_pr(self, url: str) -> tuple[str | None, str | None, str | None]:
        try:
            parsed = urlparse(url)
            parts = [p for p in parsed.path.split('/') if p]
            # Expect: owner/repo/pull/number
            if len(parts) >= 4 and parts[2] == 'pull':
                return parts[0], parts[1], parts[3]
        except Exception:
            pass
        return None, None, None
    
    async def validate_github_urls(self, manifest_data: dict[str, Any], app_id: str) -> list[LintingError]:
        """
        Validate all GitHub URLs in a manifest.
        
        Args:
            manifest_data: Parsed manifest data
            app_id: ID of the app being validated
            
        Returns:
            List of linting errors
        """
        errors = []
        
        # Check submission URL (pull request)
        submission_url = manifest_data.get("submission")
        if submission_url:
            pr_errors = await self.validate_pull_request(str(submission_url), app_id)
            errors.extend(pr_errors)
        
        # Check repo URL (if it's a GitHub URL)
        repo_url = manifest_data.get("repo")
        if repo_url and isinstance(repo_url, str) and repo_url.strip():
            repo_errors = await self._validate_github_repo_url(str(repo_url), app_id)
            errors.extend(repo_errors)
        
        return errors
    
    async def _validate_github_repo_url(self, repo_url: str, app_id: str) -> list[LintingError]:
        """Validate a GitHub repository URL."""
        errors = []
        
        # Check if it's a valid GitHub repo URL
        if not self._is_github_repo_url(repo_url):
            errors.append(LintingError(
                id="invalid_repo_url",
                severity=Severity.WARNING,
                title=f'Invalid repository URL "{repo_url}"',
                message="Repository URL should be a valid GitHub repository URL (e.g., https://github.com/owner/repo)",
                file=f"{app_id}/umbrel-app.yml",
                properties_path="repo",
            ))
            return errors
        
        # Check if repo exists
        if not await self._repo_exists(repo_url):
            errors.append(LintingError(
                id="invalid_repo_url",
                severity=Severity.WARNING,
                title=f'Repository not found "{repo_url}"',
                message="The specified repository does not exist or is not accessible. Please check the URL.",
                file=f"{app_id}/umbrel-app.yml",
                properties_path="repo",
            ))
        
        return errors
    
    def _is_github_repo_url(self, url: str) -> bool:
        """Check if URL is a valid GitHub repository URL."""
        try:
            parsed = urlparse(url)
            
            # Must be GitHub domain
            if parsed.hostname not in ["github.com", "www.github.com"]:
                return False
            
            # Must be HTTPS
            if parsed.scheme != "https":
                return False
            
            # Must match GitHub repo path pattern: /owner/repo
            path_pattern = r"^/[^/]+/[^/]+/?$"
            return bool(re.match(path_pattern, parsed.path))
            
        except Exception:
            return False
    
    async def _repo_exists(self, repo_url: str) -> bool:
        """Check if a GitHub repository exists and is accessible."""
        if repo_url in self._cache:
            return self._cache[repo_url]
        
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                # Add headers to avoid rate limiting
                headers = {
                    "User-Agent": "umbrel-linter/1.0",
                    "Accept": "application/vnd.github.v3+json",
                }
                
                response = await client.get(repo_url, headers=headers)
                
                # Repo exists if we get 200, 404 means it doesn't exist
                exists = response.status_code == 200
                self._cache[repo_url] = exists
                return exists
                
        except Exception:
            # If we can't check (network error, etc.), assume it exists to avoid false positives
            self._cache[repo_url] = True
            return True



"""
Docker image validation utilities.

This module provides comprehensive validation for Docker images,
including registry API calls and architecture checking.
"""

from __future__ import annotations

import re
from typing import Any
import httpx
from urllib.parse import urlparse

from ..core.models import LintingError, Severity


class DockerImage:
    """Represents a Docker image with parsing and validation capabilities."""
    
    def __init__(self, host: str, path: str, tag: str | None = None, digest: str | None = None):
        """Initialize a Docker image."""
        self.host = host
        self.path = path
        self.tag = tag
        self.digest = digest
    
    @classmethod
    def from_string(cls, image_string: str) -> DockerImage:
        """Parse a Docker image string into components."""
        # Handle different image formats
        # Format: [registry/]namespace/image:tag@digest
        # Examples:
        # - docker.io/library/nginx:latest
        # - nginx:latest
        # - nginx:latest@sha256:abc123
        # - registry.example.com/myapp:1.0.0@sha256:abc123
        
        # Check for digest
        if "@" in image_string:
            image_part, digest = image_string.rsplit("@", 1)
        else:
            image_part = image_string
            digest = None
        
        # Check for tag
        if ":" in image_part:
            name_part, tag = image_part.rsplit(":", 1)
        else:
            name_part = image_part
            tag = None
        
        # Parse registry and path
        parts = name_part.split("/")
        if len(parts) == 1:
            # Single name like "nginx" -> docker.io/library/nginx
            host = "docker.io"
            path = f"library/{parts[0]}"
        elif len(parts) == 2:
            # Two parts like "nginx/nginx" or "registry.com/app"
            if "." in parts[0] or ":" in parts[0]:
                # First part looks like a registry
                host = parts[0]
                path = parts[1]
            else:
                # Two-part namespace
                host = "docker.io"
                path = "/".join(parts)
        else:
            # Multiple parts like "registry.com/namespace/app"
            host = parts[0]
            path = "/".join(parts[1:])
        
        return cls(host=host, path=path, tag=tag, digest=digest)
    
    def __str__(self) -> str:
        """String representation of the image."""
        result = f"{self.host}/{self.path}"
        if self.tag:
            result += f":{self.tag}"
        if self.digest:
            result += f"@{self.digest}"
        return result
    
    @property
    def api_host(self) -> str:
        """Get the API host for registry calls."""
        # Docker Hub API is served from registry-1.docker.io
        if self.host in {"docker.io", "www.docker.com"}:
            return "registry-1.docker.io"
        return self.host
    
    @property
    def api_path(self) -> str:
        """Get the API path for registry calls."""
        return self.path


class DockerRegistryClient:
    """Client for Docker registry API calls."""
    
    def __init__(self, timeout: float = 1.0):
        """Initialize the registry client."""
        self.timeout = timeout
        self._cache: dict[str, bool] = {}
    
    async def is_registry(self, host: str) -> bool:
        """Check if a host is a valid Docker registry."""
        if host in self._cache:
            return self._cache[host]
        
        # Special cases for well-known registries
        if host in {"docker.io", "registry-1.docker.io", "ghcr.io"}:
            self._cache[host] = True
            return True
        
        try:
            async with httpx.AsyncClient(timeout=self.timeout, follow_redirects=True) as client:
                response = await client.get(f"https://{host}/v2/")
                # Per Docker Registry HTTP API V2 spec, registries may return 200 (OK)
                # or 401 (Unauthorized) for /v2/ endpoint, while still being valid.
                is_registry = False
                if response.status_code in (200, 401):
                    is_registry = True
                # Some registries set explicit header
                if response.headers.get("Docker-Distribution-Api-Version") == "registry/2.0":
                    is_registry = True
                self._cache[host] = is_registry
                return is_registry
        except Exception:
            self._cache[host] = False
            return False
    
    async def get_architectures(self, image: DockerImage) -> list[dict[str, str]]:
        """Get supported architectures for a Docker image."""
        try:
            async with httpx.AsyncClient(timeout=self.timeout, follow_redirects=True) as client:
                # Get manifest
                ref = image.digest or image.tag or "latest"
                manifest_url = f"https://{image.api_host}/v2/{image.api_path}/manifests/{ref}"
                accept = (
                    "application/vnd.oci.image.manifest.v1+json, "
                    "application/vnd.oci.image.index.v1+json, "
                    "application/vnd.docker.distribution.manifest.v2+json, "
                    "application/vnd.docker.distribution.manifest.list.v2+json"
                )
                headers = {"Accept": accept}

                response = await client.get(manifest_url, headers=headers)

                # Handle auth challenge (e.g., ghcr.io returns 401 with WWW-Authenticate)
                if response.status_code == 401:
                    www_auth = response.headers.get("WWW-Authenticate", "")
                    params = self._parse_www_authenticate(www_auth)
                    realm = params.get("realm")
                    if realm:
                        # scope: repository:<name>:pull
                        scope = params.get("scope") or f"repository:{image.api_path}:pull"
                        service = params.get("service")
                        token_params = {"scope": scope}
                        if service:
                            token_params["service"] = service
                        token_resp = await client.get(realm, params=token_params, headers={"Accept": "application/json"})
                        if token_resp.status_code == 200:
                            token = token_resp.json().get("token") or token_resp.json().get("access_token")
                            if token:
                                headers["Authorization"] = f"Bearer {token}"
                                response = await client.get(manifest_url, headers=headers)

                response.raise_for_status()

                manifest = response.json()
                content_type = response.headers.get("Content-Type", "")

                # Parse architectures based on manifest type
                if "manifest.list" in content_type or "image.index" in content_type or manifest.get("manifests"):
                    # Multi-arch manifest
                    architectures = []
                    for manifest_ref in manifest.get("manifests", []):
                        platform = manifest_ref.get("platform", {})
                        architectures.append({
                            "os": platform.get("os", "linux"),
                            "architecture": platform.get("architecture", "amd64"),
                            "variant": platform.get("variant")
                        })
                    return architectures
                else:
                    # Single-arch manifest
                    return [{"os": "linux", "architecture": "amd64"}]

        except httpx.HTTPStatusError as e:
            # Propagate status code info up
            raise Exception(f"{e.response.status_code} {e.response.reason_phrase}")
        except Exception as e:
            raise Exception(f"Failed to get architectures for {image}: {e}")

    def _parse_www_authenticate(self, header: str) -> dict:
        params: dict[str, str] = {}
        if not header:
            return params
        # Format: Bearer realm="...",service="...",scope="..."
        try:
            parts = header.split(" ", 1)
            if len(parts) == 2:
                for item in parts[1].split(","):
                    if "=" in item:
                        k, v = item.split("=", 1)
                        params[k.strip()] = v.strip().strip('"')
        except Exception:
            pass
        return params
    
    async def validate_image(self, image: DockerImage) -> list[LintingError]:
        """Validate a Docker image and return any errors."""
        errors = []
        
        # Check if it's a valid registry
        if not await self.is_registry(image.api_host):
            errors.append(LintingError(
                id="invalid_docker_image_name",
                severity=Severity.ERROR,
                title=f'Invalid registry "{image.api_host}"',
                message=f'The registry "{image.api_host}" is not a valid Docker registry',
                file="docker-compose.yml",
                properties_path="services.image"
            ))
            return errors
        
        try:
            # Get architectures
            architectures = await self.get_architectures(image)
            
            # Check if it supports both arm64 and amd64
            has_arm64 = any(
                arch.get("os") == "linux" and arch.get("architecture") == "arm64"
                for arch in architectures
            )
            has_amd64 = any(
                arch.get("os") == "linux" and arch.get("architecture") == "amd64"
                for arch in architectures
            )
            
            if not (has_arm64 and has_amd64):
                errors.append(LintingError(
                    id="invalid_image_architectures",
                    severity=Severity.ERROR,
                    title=f'Invalid image architectures for image "{image}"',
                    message=f'The image "{image}" does not support the architectures "arm64" and "amd64". Please make sure that the image supports both architectures.',
                    file="docker-compose.yml",
                    properties_path="services.image"
                ))
                
        except Exception as e:
            message = str(e)
            # If we cannot authenticate to fetch manifest (e.g., 401), do not fail hard
            if "401" in message or "Unauthorized" in message:
                errors.append(LintingError(
                    id="image_architecture_unverified",
                    severity=Severity.INFO,
                    title=f'Could not verify architectures for image "{image}"',
                    message="The registry requires authentication to inspect the image manifest. Skipping architecture verification.",
                    file="docker-compose.yml",
                    properties_path="services.image"
                ))
            else:
                errors.append(LintingError(
                    id="invalid_docker_image_name",
                    severity=Severity.ERROR,
                    title=f'Invalid image name "{image}"',
                    message=message,
                    file="docker-compose.yml",
                    properties_path="services.image"
                ))
        
        return errors


class DockerImageValidator:
    """Validator for Docker images with registry integration."""
    
    def __init__(self):
        """Initialize the validator."""
        self.registry_client = DockerRegistryClient()
    
    async def validate_images(self, services: dict[str, Any], app_id: str, check_architectures: bool = False) -> list[LintingError]:
        """Validate Docker images in services."""
        errors = []
        
        for service_name, service_config in services.items():
            image_string = service_config.get("image")
            if not image_string:
                continue
            
            # Parse image
            try:
                image = DockerImage.from_string(image_string)
            except Exception as e:
                errors.append(LintingError(
                    id="invalid_docker_image_name",
                    severity=Severity.ERROR,
                    title=f'Invalid image name "{image_string}"',
                    message=str(e),
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.image"
                ))
                continue
            
            # Enforce image format with immutable digest: <name>:<tag>@sha256:<64-hex>
            # This avoids mutable tags and ensures reproducible pulls.
            if not re.match(r"^[a-zA-Z0-9./:_-]+:[^@]+@sha256:[0-9a-fA-F]{64}$", image_string):
                errors.append(LintingError(
                    id="invalid_docker_image_name",
                    severity=Severity.ERROR,
                    title=f'Invalid image name "{image_string}"',
                    message='Images must include an immutable digest: "<name>:<version-tag>@sha256:<64-hex>"',
                    file=f"{app_id}/docker-compose.yml",
                    properties_path=f"services.{service_name}.image"
                ))
            else:
                # Check for "latest" tag
                if image.tag == "latest":
                    errors.append(LintingError(
                        id="invalid_docker_image_name",
                        severity=Severity.WARNING,
                        title=f'Invalid image tag "{image.tag}"',
                        message='Images should not use the "latest" tag',
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.image"
                    ))
            
            # Check architectures if requested
            if check_architectures:
                try:
                    image_errors = await self.registry_client.validate_image(image)
                    # Update file paths for the specific service
                    for error in image_errors:
                        error.file = f"{app_id}/docker-compose.yml"
                        error.properties_path = f"services.{service_name}.image"
                    errors.extend(image_errors)
                except Exception as e:
                    errors.append(LintingError(
                        id="invalid_docker_image_name",
                        severity=Severity.ERROR,
                        title=f'Failed to validate image "{image_string}"',
                        message=str(e),
                        file=f"{app_id}/docker-compose.yml",
                        properties_path=f"services.{service_name}.image"
                    ))
        
        return errors



"""
YAML validation utilities.

This module provides functions for validating YAML files and extracting
source map information for error reporting.
"""

from __future__ import annotations

from typing import Any

import yaml

from ..core.models import ColumnRange, LineRange, LintingError, Severity


def parse_yaml_with_error_handling(content: str, filename: str) -> tuple[dict[str, Any] | None, LintingError | None]:
    """
    Parse YAML content with comprehensive error handling.
    
    Args:
        content: YAML content as string
        filename: Name of the file being parsed
        
    Returns:
        Tuple of (parsed_data, error). If parsing succeeds, error is None.
        If parsing fails, parsed_data is None and error contains details.
    """
    try:
        data = yaml.safe_load(content)
        return data, None
    except yaml.YAMLError as e:
        error = LintingError(
            id="invalid_yaml_syntax",
            severity=Severity.ERROR,
            title=f"{filename} is not a valid YAML file",
            message=str(e),
            file=filename,
        )
        return None, error


def get_source_map_for_path(content: str, path: list[str]) -> dict[str, LineRange | ColumnRange]:
    """
    Extract source map information for a given JSON path.
    
    This is a simplified implementation. In a real implementation, you would
    use a proper YAML parser with source map support like ruamel.yaml.
    
    Args:
        content: YAML content as string
        path: JSON path to the property
        
    Returns:
        Dictionary with line and column information
    """
    lines = content.split("\n")
    result = {}

    # Simple line-based search for the property
    for i, line in enumerate(lines):
        if any(part in line for part in path):
            result["line"] = LineRange(start=i + 1, end=i + 1)
            # Find column position (simplified)
            for j, char in enumerate(line):
                if char in [":", "="]:
                    result["column"] = ColumnRange(start=j + 1, end=j + 1)
                    break
            break

    return result


def validate_yaml_structure(data: dict[str, Any], required_fields: list[str], filename: str) -> list[LintingError]:
    """
    Validate that required fields exist in the YAML data.
    
    Args:
        data: Parsed YAML data
        required_fields: List of required field names
        filename: Name of the file being validated
        
    Returns:
        List of linting errors for missing fields
    """
    errors = []

    for field in required_fields:
        if field not in data:
            errors.append(LintingError(
                id="missing_required_field",
                severity=Severity.ERROR,
                title=f"Missing required field: {field}",
                message=f"The '{field}' field is required in {filename}",
                file=filename,
                properties_path=field,
            ))

    return errors


def validate_yaml_types(data: dict[str, Any], type_mapping: dict[str, type], filename: str) -> list[LintingError]:
    """
    Validate that fields have the correct types.
    
    Args:
        data: Parsed YAML data
        type_mapping: Dictionary mapping field names to expected types
        filename: Name of the file being validated
        
    Returns:
        List of linting errors for type mismatches
    """
    errors = []

    for field, expected_type in type_mapping.items():
        if field in data:
            value = data[field]
            if not isinstance(value, expected_type):
                errors.append(LintingError(
                    id="invalid_type",
                    severity=Severity.ERROR,
                    title=f"Invalid type for field: {field}",
                    message=f"Expected {expected_type.__name__}, got {type(value).__name__}",
                    file=filename,
                    properties_path=field,
                ))

    return errors


def validate_boolean_strings(data: dict[str, Any], filename: str) -> list[LintingError]:
    """
    Validate that boolean values are strings (for Docker Compose V1 compatibility).
    
    Args:
        data: Parsed YAML data
        filename: Name of the file being validated
        
    Returns:
        List of linting errors for boolean type issues
    """
    errors = []

    def check_dict(d: dict[str, Any], path: str = "") -> None:
        for key, value in d.items():
            current_path = f"{path}.{key}" if path else key

            if isinstance(value, dict):
                check_dict(value, current_path)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        check_dict(item, f"{current_path}[{i}]")
            elif isinstance(value, bool):
                errors.append(LintingError(
                    id="invalid_yaml_boolean_value",
                    severity=Severity.ERROR,
                    title=f"Invalid YAML boolean value for key '{key}'",
                    message=f"Boolean values should be strings like '{str(value).lower()}' instead of {value}",
                    file=filename,
                    properties_path=current_path,
                ))

    check_dict(data)
    return errors



"""
Core data models for the Umbrel linter.

This module defines the fundamental data structures used throughout the linter,
including linting results, error types, and configuration models.
"""

from __future__ import annotations

from enum import Enum

from pydantic import BaseModel, Field, validator


class Severity(str, Enum):
    """Severity levels for linting results."""

    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class LintingError(BaseModel):
    """Represents a linting error with detailed information."""

    id: str = Field(..., description="Unique identifier for the error type")
    severity: Severity = Field(..., description="Severity level of the error")
    title: str = Field(..., description="Short title describing the error")
    message: str = Field(..., description="Detailed error message")
    file: str = Field(..., description="File path where the error occurred")
    properties_path: str | None = Field(None, description="JSON path to the problematic property")
    line: LineRange | None = Field(None, description="Line range where the error occurred")
    column: ColumnRange | None = Field(None, description="Column range where the error occurred")

    class Config:
        """Pydantic configuration."""
        pass


class LineRange(BaseModel):
    """Represents a range of lines in a file."""

    start: int = Field(..., ge=1, description="Starting line number (1-indexed)")
    end: int = Field(..., ge=1, description="Ending line number (1-indexed)")

    @validator("end")
    def end_must_be_gte_start(cls, v, values):
        """Ensure end line is greater than or equal to start line."""
        if "start" in values and v < values["start"]:
            raise ValueError("end must be greater than or equal to start")
        return v


class ColumnRange(BaseModel):
    """Represents a range of columns in a file."""

    start: int = Field(..., ge=1, description="Starting column number (1-indexed)")
    end: int = Field(..., ge=1, description="Ending column number (1-indexed)")

    @validator("end")
    def end_must_be_gte_start(cls, v, values):
        """Ensure end column is greater than or equal to start column."""
        if "start" in values and v < values["start"]:
            raise ValueError("end must be greater than or equal to start")
        return v


class LintingResult(BaseModel):
    """Container for linting results."""

    errors: list[LintingError] = Field(default_factory=list, description="List of linting errors")
    success: bool = Field(True, description="Whether linting was successful")
    total_errors: int = Field(0, description="Total number of errors")
    total_warnings: int = Field(0, description="Total number of warnings")
    total_info: int = Field(0, description="Total number of info messages")

    def add_error(self, error: LintingError) -> None:
        """Add a linting error to the result."""
        self.errors.append(error)
        
        # Only set success to False for actual errors (not info messages)
        if error.severity == Severity.ERROR:
            self.success = False

        # Update counters
        if error.severity == Severity.ERROR:
            self.total_errors += 1
        elif error.severity == Severity.WARNING:
            self.total_warnings += 1
        elif error.severity == Severity.INFO:
            self.total_info += 1

    def has_errors(self) -> bool:
        """Check if there are any errors (not just warnings or info)."""
        return any(error.severity == Severity.ERROR for error in self.errors)

    def has_warnings(self) -> bool:
        """Check if there are any warnings."""
        return any(error.severity == Severity.WARNING for error in self.errors)

    def has_info(self) -> bool:
        """Check if there are any info messages."""
        return any(error.severity == Severity.INFO for error in self.errors)


class FileEntry(BaseModel):
    """Represents a file or directory entry."""

    path: str = Field(..., description="Path to the file or directory")
    type: str = Field(..., description="Type: 'file' or 'directory'")

    @validator("type")
    def type_must_be_valid(cls, v):
        """Ensure type is either 'file' or 'directory'."""
        if v not in ["file", "directory"]:
            raise ValueError('type must be either "file" or "directory"')
        return v


class LinterConfig(BaseModel):
    """Configuration for the linter."""

    check_image_architectures: bool = Field(
        default=True,
        description="Whether to check Docker image architectures",
    )
    log_level: Severity = Field(
        default=Severity.WARNING,
        description="Minimum log level to display",
    )
    strict_mode: bool = Field(
        default=False,
        description="Whether to run in strict mode (treat warnings as errors)",
    )
    ignore_patterns: list[str] = Field(
        default_factory=list,
        description="File patterns to ignore during linting",
    )

    class Config:
        """Pydantic configuration."""
        pass


class AppStoreType(str, Enum):
    """Type of app store."""

    OFFICIAL = "official"
    COMMUNITY = "community"


class LintingContext(BaseModel):
    """Context information for linting operations."""

    app_id: str | None = Field(None, description="ID of the app being linted")
    app_store_type: AppStoreType = Field(
        default=AppStoreType.COMMUNITY,
        description="Type of app store",
    )
    is_new_submission: bool = Field(
        default=False,
        description="Whether this is a new app submission",
    )
    pull_request_url: str | None = Field(
        None,
        description="URL of the pull request for new submissions",
    )
    all_app_manifests: list[str] = Field(
        default_factory=list,
        description="All app manifest contents for cross-validation",
    )

    class Config:
        """Pydantic configuration."""
        use_enum_values = True



"""
Umbrel Linter - A Python linter for Umbrel app stores and applications.

This package provides comprehensive linting capabilities for Umbrel applications,
including YAML validation, Docker Compose validation, and directory structure checks.
"""

__version__ = "0.1.0"
__author__ = "Umbrel Team"
__email__ = "linter@umbrel.com"

from .core.linter import UmbrelLinter
from .core.models import LintingError, LintingResult, Severity

__all__ = [
    "LintingError",
    "LintingResult",
    "Severity",
    "UmbrelLinter",
]



"""
Variable mocking utilities for Docker Compose files.

This module provides functionality to mock environment variables
in Docker Compose files for proper validation.
"""

from __future__ import annotations

import re
from typing import Any


class VariableMocker:
    """Mocks environment variables in Docker Compose files."""
    
    def __init__(self):
        """Initialize the variable mocker."""
        pass
    
    def mock_variables(self, content: str) -> str:
        """
        Mock environment variables in Docker Compose content.
        
        Args:
            content: Docker Compose YAML content
            
        Returns:
            Content with mocked variables
        """
        variables = self._extract_variables(content)
        mocked_variables = self._find_mocks(variables)
        return self._replace_variables(content, mocked_variables)
    
    def _extract_variables(self, content: str) -> list[dict[str, str]]:
        """Extract variables from content."""
        # Match both ${VAR} and $VAR patterns
        pattern = r'\$(?:([a-zA-Z0-9_\-:]+)|\{([a-zA-Z0-9_\-:]+)\})'
        matches = re.findall(pattern, content)
        
        variables = []
        for match in matches:
            variable = match[0] if match[0] else match[1]
            full_variable = f"${{{variable}}}" if "{" in content[content.find(f"${variable}"):content.find(f"${variable}") + len(f"${variable}") + 1] else f"${variable}"
            
            variables.append({
                "full_variable": full_variable,
                "variable": variable,
                "mock": ""
            })
        
        return variables
    
    def _find_mocks(self, variables: list[dict[str, str]]) -> list[dict[str, str]]:
        """Find appropriate mock values for variables."""
        for var in variables:
            variable = var["variable"]
            
            if "_IP" in variable:
                var["mock"] = "10.10.10.10"
            elif "_PORT" in variable:
                # Random port between 1024 and 65535 to make json schema happy
                import random
                var["mock"] = str(random.randint(1024, 65535))
            elif "_PASS" in variable:
                var["mock"] = "password"
            elif "_USER" in variable:
                var["mock"] = "username"
            elif "_DIR" in variable:
                var["mock"] = "/path/to/dir"
            elif "_PATH" in variable:
                var["mock"] = "/some/path"
            elif "_SERVICE" in variable:
                var["mock"] = "service"
            elif "_SEED" in variable:
                var["mock"] = "seed"
            elif "_CONFIG" in variable:
                var["mock"] = "/path/to/config"
            elif "_MODE" in variable:
                var["mock"] = "production"
            elif "_NETWORK" in variable:
                var["mock"] = "network"
            elif "_DOMAIN" in variable:
                var["mock"] = "domain.com"
            elif "_NAME" in variable:
                var["mock"] = "name"
            elif "_VERSION" in variable:
                var["mock"] = "1.0.0"
            elif "_ROOT" in variable:
                var["mock"] = "/path/to/root"
            elif "_KEY" in variable:
                var["mock"] = "key"
            elif "_SECRET" in variable:
                var["mock"] = "secret"
            elif "_TOKEN" in variable:
                var["mock"] = "token"
            elif "_HOST" in variable:
                var["mock"] = "host"
            else:
                var["mock"] = "mocked"
        
        return variables
    
    def _replace_variables(self, content: str, variables: list[dict[str, str]]) -> str:
        """Replace variables with their mock values."""
        for var in variables:
            content = content.replace(var["full_variable"], var["mock"])
        return content



"""
Tests for core models.

This module contains tests for the core data models used throughout the linter.
"""

import pytest
from pydantic import ValidationError

from umbrel_linter.core.models import (
    LintingError, 
    Severity, 
    LineRange, 
    ColumnRange,
    LintingResult,
    FileEntry,
    LinterConfig,
    LintingContext,
    AppStoreType
)


class TestLintingError:
    """Test LintingError model."""
    
    def test_create_linting_error(self):
        """Test creating a linting error."""
        error = LintingError(
            id="test_error",
            severity=Severity.ERROR,
            title="Test Error",
            message="This is a test error",
            file="test.yml"
        )
        
        assert error.id == "test_error"
        assert error.severity == Severity.ERROR
        assert error.title == "Test Error"
        assert error.message == "This is a test error"
        assert error.file == "test.yml"
        assert error.properties_path is None
        assert error.line is None
        assert error.column is None
    
    def test_create_linting_error_with_optional_fields(self):
        """Test creating a linting error with optional fields."""
        line_range = LineRange(start=1, end=5)
        column_range = ColumnRange(start=10, end=20)
        
        error = LintingError(
            id="test_error",
            severity=Severity.WARNING,
            title="Test Warning",
            message="This is a test warning",
            file="test.yml",
            properties_path="services.app",
            line=line_range,
            column=column_range
        )
        
        assert error.properties_path == "services.app"
        assert error.line == line_range
        assert error.column == column_range


class TestLineRange:
    """Test LineRange model."""
    
    def test_create_line_range(self):
        """Test creating a line range."""
        line_range = LineRange(start=1, end=5)
        assert line_range.start == 1
        assert line_range.end == 5
    
    def test_line_range_validation(self):
        """Test line range validation."""
        # Valid range
        line_range = LineRange(start=1, end=5)
        assert line_range.start == 1
        assert line_range.end == 5
        
        # Same start and end
        line_range = LineRange(start=3, end=3)
        assert line_range.start == 3
        assert line_range.end == 3
    
    def test_line_range_invalid_end(self):
        """Test line range with invalid end."""
        with pytest.raises(ValidationError):
            LineRange(start=5, end=1)  # end < start


class TestColumnRange:
    """Test ColumnRange model."""
    
    def test_create_column_range(self):
        """Test creating a column range."""
        column_range = ColumnRange(start=10, end=20)
        assert column_range.start == 10
        assert column_range.end == 20
    
    def test_column_range_validation(self):
        """Test column range validation."""
        # Valid range
        column_range = ColumnRange(start=10, end=20)
        assert column_range.start == 10
        assert column_range.end == 20
        
        # Same start and end
        column_range = ColumnRange(start=15, end=15)
        assert column_range.start == 15
        assert column_range.end == 15
    
    def test_column_range_invalid_end(self):
        """Test column range with invalid end."""
        with pytest.raises(ValidationError):
            ColumnRange(start=20, end=10)  # end < start


class TestLintingResult:
    """Test LintingResult model."""
    
    def test_create_empty_result(self):
        """Test creating an empty result."""
        result = LintingResult()
        assert result.errors == []
        assert result.success is True
        assert result.total_errors == 0
        assert result.total_warnings == 0
        assert result.total_info == 0
    
    def test_add_error(self):
        """Test adding an error to result."""
        result = LintingResult()
        error = LintingError(
            id="test_error",
            severity=Severity.ERROR,
            title="Test Error",
            message="This is a test error",
            file="test.yml"
        )
        
        result.add_error(error)
        
        assert len(result.errors) == 1
        assert result.errors[0] == error
        assert result.success is False
        assert result.total_errors == 1
        assert result.total_warnings == 0
        assert result.total_info == 0
    
    def test_add_warning(self):
        """Test adding a warning to result."""
        result = LintingResult()
        warning = LintingError(
            id="test_warning",
            severity=Severity.WARNING,
            title="Test Warning",
            message="This is a test warning",
            file="test.yml"
        )
        
        result.add_error(warning)
        
        assert len(result.errors) == 1
        assert result.success is False
        assert result.total_errors == 0
        assert result.total_warnings == 1
        assert result.total_info == 0
    
    def test_add_info(self):
        """Test adding an info message to result."""
        result = LintingResult()
        info = LintingError(
            id="test_info",
            severity=Severity.INFO,
            title="Test Info",
            message="This is a test info message",
            file="test.yml"
        )
        
        result.add_error(info)
        
        assert len(result.errors) == 1
        assert result.success is False
        assert result.total_errors == 0
        assert result.total_warnings == 0
        assert result.total_info == 1
    
    def test_has_errors(self):
        """Test has_errors method."""
        result = LintingResult()
        assert not result.has_errors()
        
        # Add warning - should not count as error
        warning = LintingError(
            id="test_warning",
            severity=Severity.WARNING,
            title="Test Warning",
            message="This is a test warning",
            file="test.yml"
        )
        result.add_error(warning)
        assert not result.has_errors()
        
        # Add error - should count as error
        error = LintingError(
            id="test_error",
            severity=Severity.ERROR,
            title="Test Error",
            message="This is a test error",
            file="test.yml"
        )
        result.add_error(error)
        assert result.has_errors()
    
    def test_has_warnings(self):
        """Test has_warnings method."""
        result = LintingResult()
        assert not result.has_warnings()
        
        # Add info - should not count as warning
        info = LintingError(
            id="test_info",
            severity=Severity.INFO,
            title="Test Info",
            message="This is a test info message",
            file="test.yml"
        )
        result.add_error(info)
        assert not result.has_warnings()
        
        # Add warning - should count as warning
        warning = LintingError(
            id="test_warning",
            severity=Severity.WARNING,
            title="Test Warning",
            message="This is a test warning",
            file="test.yml"
        )
        result.add_error(warning)
        assert result.has_warnings()
    
    def test_has_info(self):
        """Test has_info method."""
        result = LintingResult()
        assert not result.has_info()
        
        # Add error - should not count as info
        error = LintingError(
            id="test_error",
            severity=Severity.ERROR,
            title="Test Error",
            message="This is a test error",
            file="test.yml"
        )
        result.add_error(error)
        assert not result.has_info()
        
        # Add info - should count as info
        info = LintingError(
            id="test_info",
            severity=Severity.INFO,
            title="Test Info",
            message="This is a test info message",
            file="test.yml"
        )
        result.add_error(info)
        assert result.has_info()


class TestFileEntry:
    """Test FileEntry model."""
    
    def test_create_file_entry(self):
        """Test creating a file entry."""
        entry = FileEntry(path="test.yml", type="file")
        assert entry.path == "test.yml"
        assert entry.type == "file"
    
    def test_create_directory_entry(self):
        """Test creating a directory entry."""
        entry = FileEntry(path="test_dir", type="directory")
        assert entry.path == "test_dir"
        assert entry.type == "directory"
    
    def test_invalid_type(self):
        """Test invalid type validation."""
        with pytest.raises(ValidationError):
            FileEntry(path="test.yml", type="invalid")


class TestLinterConfig:
    """Test LinterConfig model."""
    
    def test_default_config(self):
        """Test default configuration."""
        config = LinterConfig()
        assert config.check_image_architectures is False
        assert config.log_level == Severity.WARNING
        assert config.strict_mode is False
        assert config.ignore_patterns == []
    
    def test_custom_config(self):
        """Test custom configuration."""
        config = LinterConfig(
            check_image_architectures=True,
            log_level=Severity.ERROR,
            strict_mode=True,
            ignore_patterns=["*.tmp", "*.log"]
        )
        assert config.check_image_architectures is True
        assert config.log_level == Severity.ERROR
        assert config.strict_mode is True
        assert config.ignore_patterns == ["*.tmp", "*.log"]


class TestLintingContext:
    """Test LintingContext model."""
    
    def test_default_context(self):
        """Test default context."""
        context = LintingContext()
        assert context.app_id is None
        assert context.app_store_type == AppStoreType.COMMUNITY
        assert context.is_new_submission is False
        assert context.pull_request_url is None
        assert context.all_app_manifests == []
    
    def test_custom_context(self):
        """Test custom context."""
        context = LintingContext(
            app_id="test-app",
            app_store_type=AppStoreType.OFFICIAL,
            is_new_submission=True,
            pull_request_url="https://github.com/user/repo/pull/123",
            all_app_manifests=["manifest1", "manifest2"]
        )
        assert context.app_id == "test-app"
        assert context.app_store_type == AppStoreType.OFFICIAL
        assert context.is_new_submission is True
        assert context.pull_request_url == "https://github.com/user/repo/pull/123"
        assert context.all_app_manifests == ["manifest1", "manifest2"]
